/*
 * Zynaptic Stash - An asynchronous persistence framework for Java.
 * 
 * Copyright (c) 2009-2015, Zynaptic Limited. All rights reserved. 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 only, as published by
 * the Free Software Foundation. Zynaptic Limited designates this particular
 * file as subject to the "Classpath" exception as provided in the LICENSE
 * file that accompanied this code.
 * 
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 * 
 * You should have received a copy of the GNU General Public License version 2
 * along with this work; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Please visit www.zynaptic.com or contact reaction@zynaptic.com if you need
 * additional information or have any questions.
 */

package com.zynaptic.stash;

import java.io.Serializable;

import com.zynaptic.reaction.Deferred;

/**
 * Provides the interface for manipulating an individual raw Java object in the
 * persistent stash storage.
 * 
 * @param <T>
 *          This is the generic type specifier which determines the class of
 *          objects which may be stored in the stash using the stash entry
 *          handle.
 * 
 * @author Chris Holgate
 */
public interface StashObjectEntry<T extends Serializable> {

  /**
   * Accesses the unique stash identifier which is associated with the stash
   * object entry.
   * 
   * @return Returns the unique stash identifier for the stash object entry.
   */
  public String getStashId();

  /**
   * Accesses the human readable stash entry name which is associated with the
   * stash object entry.
   * 
   * @return Returns the human readable stash entry name for the stash object
   *         entry.
   */
  public String getStashName();

  /**
   * Accesses the default value for the stash object entry. This is the default
   * value which was specified when the stash entry was created and is used for
   * run time type checking.
   * 
   * @return Returns the default value for the stash object entry.
   */
  public T getDefaultValue();

  /**
   * Accesses the current stored value for the stash object entry. This is the
   * most recent value which was committed to the persistent store.
   * 
   * @return Returns a deferred event object which will have its callbacks
   *         executed on completion. It will pass the current value for the
   *         stash object entry as the callback parameter.
   */
  public Deferred<T> getCurrentValue();

  /**
   * Assigns a new value to the stash object entry. This will commit the
   * specified data to the persistent store.
   * 
   * @param newValue
   *          This is the new value for the object entry which is to be
   *          committed to the stash.
   * @return Returns a deferred event object which will have its callbacks
   *         executed once the data has been committed to the persistent store.
   *         It will pass a boolean value which will normally be set to 'true'
   *         as the callback parameter.
   */
  public Deferred<Boolean> setCurrentValue(T newValue);

}
